import collections
from copy import copy
import itertools as it

import alifedata_phyloinformatics_convert as apc
import numpy as np
import pandas as pd
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import omit_last


def _compare_compiled_phylogenies(
    control_phylogeny_df: pd.DataFrame,
    test_phylogeny_df: pd.DataFrame,
    share_common_ancestor: bool,
) -> None:
    # compile tree tracked with handle tracker
    control_trees = apc.alife_dataframe_to_dendropy_trees(
        control_phylogeny_df,
    )
    if share_common_ancestor:
        assert len(control_trees) == 1

    # compile tree tracked with gc tracker
    assert len(test_phylogeny_df) == len(test_phylogeny_df["id"].unique())

    test_trees = apc.alife_dataframe_to_dendropy_trees(
        test_phylogeny_df,
    )
    assert sum(
        1 for test_tree in test_trees for __ in test_tree.leaf_node_iter()
    ) == sum(
        1
        for control_tree in control_trees
        for __ in control_tree.leaf_node_iter()
    )
    assert (
        len(
            set(
                node.level()
                for test_tree in control_trees
                for node in test_tree.leaf_node_iter()
            )
        )
        == 1
    )
    assert all(
        node.level() == node.origin_time
        for test_tree in test_trees
        for node in test_tree
    )

    assert len(test_trees) == len(control_trees)

    # setup taxa in both trees
    for trees in test_trees, control_trees:
        for i, tree in enumerate(trees):
            for j, n in enumerate(tree):
                n.taxon = tree.taxon_namespace.new_taxon(f"{i} {j}")

    # check phylogenetic trees with and without test are identical
    assert collections.Counter(
        node.level() for control_tree in control_trees for node in control_tree
    ) == collections.Counter(
        node.level() for test_tree in test_trees for node in test_tree
    )
    assert collections.Counter(
        control_tree.mrca(taxa=[n1.taxon, n2.taxon]).level()
        for control_tree in control_trees
        for n1, n2 in it.combinations(control_tree.leaf_node_iter(), 2)
    ) == collections.Counter(
        test_tree.mrca(taxa=[n1.taxon, n2.taxon]).level()
        for test_tree in test_trees
        for n1, n2 in it.combinations(test_tree.leaf_node_iter(), 2)
    )


@pytest.mark.parametrize(
    "population_size",
    [
        20,
        100,
    ],
)
@pytest.mark.parametrize(
    "share_common_ancestor",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "working_buffer_size_mx",
    [
        1,
        10,
    ],
)
def test_GarbageCollectingPhyloTracker(
    population_size, share_common_ancestor, working_buffer_size_mx
):

    # setup population and perfect trackers
    common_ancestor = hstrat.PerfectBacktrackHandle()
    handle_population = (
        [common_ancestor.CreateDescendant() for __ in range(population_size)]
        if share_common_ancestor
        else [hstrat.PerfectBacktrackHandle() for __ in range(population_size)]
    )
    tracker = hstrat.GarbageCollectingPhyloTracker(
        initial_population=population_size,
        share_common_ancestor=share_common_ancestor,
        working_buffer_size=population_size * working_buffer_size_mx + 1,
    )

    # evolve fixed-size population with random selection
    for epoch_length in 0, 3, 50, 15, 5:
        for _generation in range(epoch_length):
            _ = _generation
            parent_locs = np.random.randint(
                population_size, size=population_size
            )
            tracker.ElapseGeneration(parent_locs)
            handle_population = [
                handle_population[loc].CreateDescendant()
                for loc in parent_locs
            ]

        _compare_compiled_phylogenies(
            hstrat.compile_perfect_backtrack_phylogeny(handle_population),
            tracker.CompilePhylogeny(),
            share_common_ancestor,
        )


def test_GarbateCollectingPhyloTracker_ApplyLocSwaps():

    # setup population and perfect trackers
    population_size = 4
    common_ancestor = hstrat.PerfectBacktrackHandle(data=0)
    handle_population = [
        common_ancestor.CreateDescendant(data=loc)
        for loc in range(population_size)
    ]
    tracker = hstrat.GarbageCollectingPhyloTracker(
        initial_population=population_size,
        working_buffer_size=21,
    )

    # evolve fixed-size population with random selection
    for _generation in range(50):
        _ = _generation
        parent_locs = np.random.randint(population_size, size=population_size)
        tracker.ElapseGeneration(parent_locs)
        handle_population = [
            handle_population[loc].CreateDescendant() for loc in parent_locs
        ]

        swapfrom_locs = np.random.randint(
            population_size, size=population_size
        )
        swapto_locs = np.random.randint(population_size, size=population_size)

        tracker.ApplyLocSwaps(swapfrom_locs, swapto_locs)
        for from_loc, to_loc in zip(swapfrom_locs, swapto_locs):
            handle_population[from_loc], handle_population[to_loc] = (
                handle_population[to_loc],
                handle_population[from_loc],
            )

        for loc, handle in enumerate(handle_population):
            handle.data = loc

    handle_alife_df = hstrat.compile_perfect_backtrack_phylogeny(
        handle_population
    )
    handle_tree = apc.alife_dataframe_to_dendropy_tree(
        handle_alife_df,
        setattrs={"data": "loc"},
    )
    handle_loc_lineages = {
        tuple(ancestor.loc for ancestor in leaf.ancestor_iter(inclusive=True))
        for leaf in handle_tree.leaf_node_iter()
    }

    gc_alife_df = tracker.CompilePhylogeny()
    gc_tree = apc.alife_dataframe_to_dendropy_tree(
        gc_alife_df,
        setattrs=["loc"],
    )
    gc_loc_lineages = {
        tuple(ancestor.loc for ancestor in leaf.ancestor_iter(inclusive=True))
        for leaf in gc_tree.leaf_node_iter()
    }

    assert len(handle_tree) == len(gc_tree)
    assert gc_loc_lineages == handle_loc_lineages


def test_GarbateCollectingPhyloTracker_ApplyLocPasteovers():

    # setup population and perfect trackers
    population_size = 10
    common_ancestor = hstrat.PerfectBacktrackHandle(
        data={"loc": 0, "trait": np.nan}
    )
    handle_naivepaste_population = [
        common_ancestor.CreateDescendant(
            data={"loc": loc, "trait": float(loc)}
        )
        for loc in range(population_size)
    ]
    handle_copypaste_population = [
        common_ancestor.CreateDescendant(
            data={"loc": loc, "trait": float(loc)}
        )
        for loc in range(population_size)
    ]
    tracker = hstrat.GarbageCollectingPhyloTracker(
        initial_population=np.arange(population_size),
        working_buffer_size=21,
    )

    # evolve fixed-size population with random selection
    for _generation in range(50):
        _ = _generation
        parent_locs = np.random.randint(population_size, size=population_size)
        tracker.ElapseGeneration(
            parent_locs, traits=np.arange(population_size)
        )
        handle_naivepaste_population = [
            handle_naivepaste_population[parent_loc].CreateDescendant(
                data={"trait": float(loc)}
            )
            for loc, parent_loc in enumerate(parent_locs)
        ]
        handle_copypaste_population = [
            handle_copypaste_population[parent_loc].CreateDescendant(
                data={"trait": float(loc)}
            )
            for loc, parent_loc in enumerate(parent_locs)
        ]

        # apply pasteovers
        copyfrom_locs = np.random.randint(
            population_size, size=np.random.randint(population_size)
        )
        copyto_locs = np.arange(population_size)
        np.random.shuffle(copyto_locs)
        copyto_locs = copyto_locs[: len(copyfrom_locs)]

        tracker.ApplyLocPasteovers(copyfrom_locs, copyto_locs)

        for from_loc, to_loc in zip(copyfrom_locs, copyto_locs):
            handle_naivepaste_population[to_loc] = copy(
                handle_naivepaste_population[from_loc]
            )
            handle_naivepaste_population[to_loc].data = copy(
                handle_naivepaste_population[to_loc].data
            )

        handle_copypaste_population_ = copy(handle_copypaste_population)
        for from_loc, to_loc in zip(copyfrom_locs, copyto_locs):
            handle_copypaste_population[to_loc] = copy(
                handle_copypaste_population_[from_loc]
            )
            handle_copypaste_population[to_loc].data = copy(
                handle_copypaste_population[to_loc].data
            )

        # set loc after pasteovers
        for pop in handle_naivepaste_population, handle_copypaste_population:
            for loc, handle in enumerate(pop):
                handle.data["loc"] = loc

    handle_naivepaste_tree = apc.alife_dataframe_to_dendropy_tree(
        hstrat.compile_perfect_backtrack_phylogeny(
            handle_naivepaste_population
        ),
        setattrs=["loc", "trait"],
    )
    handle_naivepaste_data_lineages = {
        tuple(
            (ancestor.loc, ancestor.trait)
            for ancestor in omit_last(leaf.ancestor_iter(inclusive=True))
        )
        for leaf in handle_naivepaste_tree.leaf_node_iter()
    }

    handle_copypaste_tree = apc.alife_dataframe_to_dendropy_tree(
        hstrat.compile_perfect_backtrack_phylogeny(
            handle_copypaste_population
        ),
        setattrs=["loc", "trait"],
    )
    handle_copypaste_data_lineages = {
        tuple(
            (ancestor.loc, ancestor.trait)
            for ancestor in omit_last(leaf.ancestor_iter(inclusive=True))
        )
        for leaf in handle_copypaste_tree.leaf_node_iter()
    }

    gc_alife_df = tracker.CompilePhylogeny()
    gc_tree = apc.alife_dataframe_to_dendropy_tree(
        gc_alife_df,
        setattrs=["loc", "trait"],
    )
    gc_data_lineages = {
        tuple(
            (ancestor.loc, ancestor.trait)
            for ancestor in omit_last(leaf.ancestor_iter(inclusive=True))
        )
        for leaf in gc_tree.leaf_node_iter()
    }

    assert (
        len(handle_copypaste_tree)
        == len(handle_naivepaste_tree)
        == len(gc_tree)
    )
    assert gc_data_lineages == handle_copypaste_data_lineages
    assert gc_data_lineages != handle_naivepaste_data_lineages

    assert sum(1 for __ in handle_copypaste_tree) == sum(1 for __ in gc_tree)
    assert len(handle_naivepaste_tree) == len(gc_tree)



import warnings

import pytest

from hstrat._auxiliary_lib import cast_int_lossless


@pytest.mark.parametrize("action", ["raise", "warn"])
def test_cast_int_lossless_with_int(action: str):
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert cast_int_lossless(-42, action) == -42
        assert cast_int_lossless(0, action, context="foo") == 0
        assert cast_int_lossless(10, action) == 10


@pytest.mark.parametrize("action", ["raise", "warn"])
def test_cast_int_lossless_with_float_no_loss(action: str):
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert cast_int_lossless(-42.0, action) == -42
        assert cast_int_lossless(0.0, action, context="foo") == 0
        assert cast_int_lossless(10.0, action) == 10


def test_cast_int_lossless_with_loss_warn():
    with pytest.warns(UserWarning):
        assert cast_int_lossless(10.1, "warn") == 10
        assert cast_int_lossless(10.1, "warn", context="foo") == 10


def test_cast_int_lossless_with_loss_raise():
    with pytest.raises(ValueError):
        cast_int_lossless(10.1, "raise")


def test_cast_int_lossless_unrecognized_action():
    with pytest.raises(RuntimeError):
        cast_int_lossless(10.1, "invalid_action")



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import pick_policy


class CalcMrcaUncertaintyAbsExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcMrcaUncertaintyAbsExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsExact",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> int:
        """Exactly how much uncertainty to estimate rank of MRCA?"""
        return pick_policy(
            policy.GetSpec()._size_curb,
            max(
                first_num_strata_deposited,
                second_num_strata_deposited,
            ),
        ).CalcMrcaUncertaintyAbsExact(
            first_num_strata_deposited,
            second_num_strata_deposited,
            actual_rank_of_mrca,
        )



import more_itertools as mit

from hstrat._auxiliary_lib import log_memory_usage


def test_log_memory_usage():
    message = []
    log_memory_usage(message.append)
    assert "available" in mit.one(message)



from ..._impl import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank as CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank_,
)


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank(
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank_
):
    pass



import numbers
import typing
import warnings

import numpy as np


def cast_int_lossless(
    value: numbers.Real,
    action: typing.Literal["warn", "raise"],
    context: str = "",
) -> int:
    """Cast a real number to an integer, with options to warn or raise an
    exception if the conversion involves precision loss.

    Parameters
    ----------
    value : numbers.Real
        Real number to be converted to an integer.
    action : {'warn', 'raise'}
        Action to take if the conversion results in precision loss.
    context : str, optional
        Additional context to include in the warning or error message.

    Returns
    -------
    int
        The integer value after conversion.

    Raises
    ------
    ValueError
        If `action` is 'raise' and the conversion results in precision loss.
    RuntimeError
        If `action` is unrecognized.

    Examples
    --------
    >>> coerce_int_warn_lossy(3.14, 'warn', 'example')
    UserWarning: Precision loss in example conversion of 3.14 to int.
    3

    >>> coerce_int_warn_lossy(3.14, 'raise')
    ValueError: Precision loss in conversion of 3.14 to int.
    """

    if not np.isclose(value, int(value)):
        message = (
            "Precision loss in"
            + f" {context}".rstrip()
            + f" conversion of {value} to int."
        )
        if action == "warn":
            warnings.warn(message)
        elif action == "raise":
            raise ValueError(message)
        else:
            raise RuntimeError(f"action {action} is unrecognized")

    return int(value)



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_collapse_trunk_asexual,
    alifestd_validate,
)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_no_trunk_column(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
        }
    )
    with pytest.raises(ValueError):
        alifestd_collapse_trunk_asexual(df, mutate=mutate)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_single_trunk(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[]"],
            "is_trunk": [True],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df, mutate=mutate)
    assert alifestd_validate(result)
    pd.testing.assert_frame_equal(result[df.columns], df)

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_no_collapse_needed1(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 8],
            "ancestor_id": [0, 0],
            "ancestor_list": ["[none]", "[0]"],
            "is_trunk": [True, False],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df)
    assert alifestd_validate(result)
    pd.testing.assert_frame_equal(
        result[df.columns].reset_index(drop=True), df
    )

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_no_collapse_needed2(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[none]", "[0]"],
            "is_trunk": [True, True],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "is_trunk": [True],
        }
    )
    pd.testing.assert_frame_equal(result[expected.columns], expected)

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_no_collapse_needed3(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 1],
            "ancestor_list": ["[none]", "[0]", "[1]"],
            "is_trunk": [False, False, False],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df)
    assert alifestd_validate(result)
    pd.testing.assert_frame_equal(result[df.columns], df)

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_collapse1(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4, 3, 5],
            "ancestor_id": [0, 0, 1, 2, 0, 5],
            "ancestor_list": ["[none]", "[0]", "[1]", "[2]", "[0]", "[none]"],
            "is_trunk": [True, True, True, False, False, False],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [0, 4, 3, 5],
            "ancestor_id": [0, 0, 0, 5],
            "ancestor_list": ["[none]", "[0]", "[0]", "[none]"],
            "is_trunk": [True, False, False, False],
        }
    )
    pd.testing.assert_frame_equal(result[expected.columns], expected)

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_collapse2(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[0]",
                "[0]",
                "[2]",
                "[3]",
            ],
            "is_trunk": [True, True, True, False, True, True, False],
        }
    )
    original_df = df.copy()
    result = alifestd_collapse_trunk_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [0, 3, 6],
            "ancestor_list": ["[none]", "[0]", "[3]"],
            "is_trunk": [True, False, False],
        }
    )
    pd.testing.assert_frame_equal(result[expected.columns], expected)

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_collapse_trunk_asexual_noncontiguous_trunk(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "is_trunk": [True, False, True],
        }
    )
    original_df = df.copy()
    with pytest.raises(ValueError):
        alifestd_collapse_trunk_asexual(df)

    if not mutate:
        assert df.equals(original_df)



import polars as pl


def get_sole_scalar_value_polars(df: pl.DataFrame, col_name: str) -> int:
    """Get scalar value from DataFrame, ensuring it is unique."""
    if df.lazy().limit(1).collect().is_empty():
        raise ValueError(f"DataFrame is empty, cannot get {col_name}")
    if (
        not df.lazy()
        .filter(pl.col(col_name).diff() != pl.lit(0))
        .limit(1)
        .collect()
        .is_empty()
    ):
        raise ValueError(f"multiple {col_name} values detected")
    return df.lazy().select(col_name).limit(1).collect().item()



import functools
import itertools as it
import os
import random

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest
from tqdm import tqdm

from hstrat import hstrat
from hstrat._auxiliary_lib import alifestd_find_leaf_ids

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "always_store_rank_in_stratum",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "assign_generation__drop_origin_time",
    [[False, False], [False, True], [True, True]],
)
@pytest.mark.parametrize(
    "num_predeposits",
    [
        0,
        1,
        10,
        100,
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.stochastic_algo.Policy(retention_probability=0.05),
        hstrat.depth_proportional_resolution_algo.Policy(
            depth_proportional_resolution=10
        ),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_descend_template_phylogeny(
    always_store_rank_in_stratum,
    assign_generation__drop_origin_time,
    num_predeposits,
    retention_policy,
    phylogeny_df,
):

    assign_generation, drop_origin_time = assign_generation__drop_origin_time

    phylogeny_df = phylogeny_df.copy()

    if assign_generation:
        phylogeny_df["generation"] = phylogeny_df["origin_time"]

    tree = apc.alife_dataframe_to_dendropy_tree(
        phylogeny_df,
        setup_edge_lengths=(not drop_origin_time) or assign_generation,
    )

    if drop_origin_time:
        phylogeny_df.drop(columns="origin_time", inplace=True)

    if drop_origin_time and not assign_generation:
        for node in tree:
            node.edge_length = 1

    tree.seed_node.edge_length = num_predeposits

    for node in tree:
        node.taxon = tree.taxon_namespace.new_taxon(label=node.id)

    tree.update_bipartitions(
        suppress_unifurcations=False,
        collapse_unrooted_basal_bifurcation=False,
    )

    # setup seed column
    seed_column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    seed_column.DepositStrata(num_stratum_depositions=num_predeposits)

    sampled_tree_nodes = random.sample([*tree], 7)
    id_index = phylogeny_df.set_index("id").index
    sorted_leaf_nodes = sorted(
        tree.leaf_node_iter(),
        key=lambda node: id_index.get_loc(node.id),
    )
    assert [n.id for n in sorted_leaf_nodes] == alifestd_find_leaf_ids(
        phylogeny_df
    )
    for extant_ids, sorted_extant_nodes in (
        (None, sorted_leaf_nodes),
        (map(lambda node: node.id, sampled_tree_nodes), sampled_tree_nodes),
    ):

        extant_population = hstrat.descend_template_phylogeny_alifestd(
            phylogeny_df,
            seed_column=seed_column,
            extant_ids=extant_ids,
            progress_wrap=functools.partial(tqdm, disable=True),
        )

        num_tips = len(sorted_extant_nodes)
        assert num_tips == len(extant_population)

        tip_depths = [
            int(tip_node.distance_from_root())
            for tip_node in sorted_extant_nodes
        ]
        assert tip_depths == [
            column.GetNumStrataDeposited() - 1 for column in extant_population
        ]

        sampled_product = it.permutations(
            random.sample(
                [*zip(extant_population, sorted_extant_nodes)],
                min(10, len(extant_population)),
            ),
            2,
        )
        spliced_product = it.permutations(
            it.islice(zip(extant_population, sorted_extant_nodes), 10),
            2,
        )

        for (c1, n1), (c2, n2) in it.chain(sampled_product, spliced_product):
            lb, ub = hstrat.calc_rank_of_mrca_bounds_between(
                c1, c2, prior="arbitrary"
            )
            assert n1 != n2
            if n1 in n2.ancestor_iter():
                mrca = n1
            elif n2 in n1.ancestor_iter():
                mrca = n2
            else:

                # dendropy mrca lookup requires leaf nodes
                while n1.num_child_nodes():
                    n1, *__ = n1.child_nodes()
                while n2.num_child_nodes():
                    n2, *__ = n2.child_nodes()

                mrca = tree.mrca(
                    taxa=[n1.taxon, n2.taxon],
                    is_bipartitions_updated=True,
                )
                # patch for dendropy bug where internal unifurcations are not accounted
                # for in mrca detection
                # see https://github.com/jeetsukumaran/DendroPy/pull/148
                while mrca.num_child_nodes() == 1:
                    (mrca,) = mrca.child_nodes()

            assert lb <= mrca.distance_from_root() < ub



import itertools as it
import os
import random

from Bio.Phylo.TreeConstruction import BaseTree
import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import networkx as nx
import pytest
from tqdm import tqdm

from hstrat import hstrat
from hstrat._auxiliary_lib import (
    alifestd_collapse_unifurcations,
    alifestd_has_multiple_roots,
    alifestd_is_chronologically_ordered,
    alifestd_validate,
    generate_n,
    random_tree,
    seed_random,
)

from . import _impl as impl

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_empty_population():
    population = []
    tree = hstrat.build_tree_trie(population)

    assert len(tree) == 0
    assert alifestd_is_chronologically_ordered(tree)
    assert alifestd_validate(tree)


def test_dual_population_no_mrca():
    organism1 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)
    organism2 = hstrat.HereditaryStratigraphicColumn().CloneNthDescendant(100)

    population = [organism1, organism2]
    names = ["foo", "bar"]

    with pytest.raises(ValueError):
        tree = hstrat.build_tree_trie(population, taxon_labels=names)

    tree = hstrat.build_tree_trie(
        population,
        taxon_labels=names,
        force_common_ancestry=True,
    )
    tree["name"] = tree["taxon_label"]
    assert not alifestd_has_multiple_roots(tree)
    assert alifestd_is_chronologically_ordered(tree)
    assert alifestd_validate(tree)

    root_clade = BaseTree.Clade(name="Inner1")
    root_clade.clades = [
        BaseTree.Clade(branch_length=100.0, name="bar"),
        BaseTree.Clade(branch_length=100.0, name="foo"),
    ]
    true_tree = BaseTree.Tree(rooted=False, root=root_clade)
    assert (
        impl.tree_unweighted_robinson_foulds_distance(
            apc.alife_dataframe_to_biopython_tree(
                alifestd_collapse_unifurcations(tree),
                setup_branch_lengths=True,
            ),
            true_tree,
        )
        == 0.0
    )


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/grandchild_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandchild_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandchild.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_aunt.newick"
        ),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtriplets_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtriplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins_and_aunt.newick"),
        impl.setup_dendropy_tree(
            f"{assets_path}/grandtwins_and_auntuncle.newick"
        ),
        impl.setup_dendropy_tree(f"{assets_path}/grandtwins.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/justroot.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/triplets.newick"),
        impl.setup_dendropy_tree(f"{assets_path}/twins.newick"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize("wrap", [lambda x: x, hstrat.col_to_specimen])
def test_handwritten_trees(orig_tree, retention_policy, wrap):
    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(10),
    )

    reconst_df = hstrat.build_tree_trie([*map(wrap, extant_population)])

    assert alifestd_validate(reconst_df)
    assert alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    reconst_tree.collapse_unweighted_edges()

    common_namespace = dp.TaxonNamespace()
    orig_tree.migrate_taxon_namespace(common_namespace)
    reconst_tree.migrate_taxon_namespace(common_namespace)

    original_distance_matrix = orig_tree.phylogenetic_distance_matrix()
    reconstructed_distance_matrix = reconst_tree.phylogenetic_distance_matrix()

    taxa = [node.taxon for node in orig_tree.leaf_node_iter()]

    for a, b in it.combinations(taxa, 2):
        assert original_distance_matrix.distance(
            a, b
        ) == reconstructed_distance_matrix.distance(a, b)


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
def test_reconstructed_mrca(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df = hstrat.build_tree_trie(extant_population)
    assert "origin_time" in reconst_df

    assert alifestd_validate(reconst_df)
    assert alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    pdm = reconst_tree.phylogenetic_distance_matrix()

    assert len(list(reconst_tree.leaf_node_iter())) == len(extant_population)
    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert {
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    } == {
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    }
    assert sorted(
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    ) == sorted(
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    )
    assert [
        leaf_node.distance_from_root() for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    for reconst_node_pair, extant_column_pair in zip(
        it.combinations(sorted_leaf_nodes, 2),
        it.combinations(extant_population, 2),
    ):
        reconst_mrca = impl.descend_unifurcations(
            pdm.mrca(*map(lambda x: x.taxon, reconst_node_pair))
        )

        (
            lower_mrca_bound,
            upper_mrca_bound,
        ) = hstrat.calc_rank_of_mrca_bounds_between(
            *extant_column_pair, prior="arbitrary"
        )

        assert (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
def test_col_specimen_consistency(orig_tree, retention_policy):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    reconst_df1 = hstrat.build_tree_trie(extant_population)
    reconst_df2 = hstrat.build_tree_trie(
        [hstrat.col_to_specimen(col) for col in extant_population],
    )

    assert reconst_df1[["id", "ancestor_list"]].equals(
        reconst_df2[["id", "ancestor_list"]]
    )


@pytest.mark.parametrize("tree_size", [10, 30, 100, 300, 1000])
@pytest.mark.parametrize(
    "differentia_width", [1, pytest.param(2, marks=pytest.mark.heavy), 8, 64]
)
@pytest.mark.parametrize(
    "tree_seed",
    [*range(10)]
    + [pytest.param(i, marks=pytest.mark.heavy) for i in range(10, 100)],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(1),
        pytest.param(
            hstrat.recency_proportional_resolution_algo.Policy(2),
            marks=pytest.mark.heavy,
        ),
        pytest.param(
            hstrat.recency_proportional_resolution_algo.Policy(10),
            marks=pytest.mark.heavy,
        ),
        hstrat.fixed_resolution_algo.Policy(2),
    ],
)
@pytest.mark.parametrize(
    "exhaustive_check",
    [pytest.param(True, marks=pytest.mark.heavy), False],
)
def test_reconstructed_mrca_fuzz(
    tree_seed, tree_size, differentia_width, retention_policy, exhaustive_check
):

    seed_random(tree_seed)
    nx_tree = random_tree(n=tree_size, seed=tree_seed, create_using=nx.DiGraph)

    extant_population = hstrat.descend_template_phylogeny_networkx(
        nx_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
            stratum_differentia_bit_width=differentia_width,
        ),
    )

    reconst_df = hstrat.build_tree_trie(
        extant_population,
        progress_wrap=tqdm,
    )
    assert "origin_time" in reconst_df

    assert alifestd_validate(reconst_df)
    assert alifestd_is_chronologically_ordered(reconst_df)
    reconst_tree = apc.alife_dataframe_to_dendropy_tree(
        reconst_df,
        setup_edge_lengths=True,
    )
    pdm = reconst_tree.phylogenetic_distance_matrix()

    assert len(list(reconst_tree.leaf_node_iter())) == len(extant_population)
    sorted_leaf_nodes = sorted(
        reconst_tree.leaf_node_iter(), key=lambda x: int(x.taxon.label)
    )
    assert {
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    } == {
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    }
    assert sorted(
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ) == sorted(
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    )
    assert [
        int(leaf_node.distance_from_root()) for leaf_node in sorted_leaf_nodes
    ] == [
        extant_col.GetNumStrataDeposited() - 1
        for extant_col in extant_population
    ]

    for reconst_node_pair, extant_column_pair, label_pair in (
        (
            (
                (sorted_leaf_nodes[first_idx], sorted_leaf_nodes[second_idx]),
                (extant_population[first_idx], extant_population[second_idx]),
                (first_idx, second_idx),
            )
            for first_idx, second_idx in generate_n(
                lambda: (
                    random.randrange(len(extant_population)),
                    random.randrange(len(extant_population)),
                ),
                10000,
            )
        )
        if not exhaustive_check
        else zip(
            it.combinations(sorted_leaf_nodes, 2),
            it.combinations(extant_population, 2),
            it.combinations(range(len(extant_population)), 2),
        )
    ):
        reconst_mrca = impl.descend_unifurcations(
            pdm.mrca(*map(lambda x: x.taxon, reconst_node_pair))
        )

        (
            lower_mrca_bound,
            upper_mrca_bound,
        ) = hstrat.calc_rank_of_mrca_bounds_between(
            *extant_column_pair,
            prior="arbitrary",
            confidence_level=0.99999,
            strict=False,
        )

        if not (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        ):
            print(reconst_mrca.taxon.label)
            print(reconst_mrca.distance_from_root())
            print()
            for i, (col, ln) in enumerate(
                zip(extant_population, sorted_leaf_nodes)
            ):
                print()
                print("label", i)
                print(ln.distance_from_root())
                print(id(col))
                print(hstrat.col_to_ascii(col))
            print(nx.forest_str(nx_tree))
            print(reconst_tree.as_ascii_plot(plot_metric="length"))

        assert (
            lower_mrca_bound
            <= reconst_mrca.distance_from_root()
            < upper_mrca_bound
        ), (
            label_pair[0],
            label_pair[1],
            id(extant_column_pair[0]),
            id(extant_column_pair[1]),
            lower_mrca_bound,
            reconst_mrca.distance_from_root(),
            upper_mrca_bound,
        )


@pytest.mark.parametrize(
    "orig_tree",
    [
        pytest.param(
            impl.setup_dendropy_tree(f"{assets_path}/nk_ecoeaselection.csv"),
            marks=pytest.mark.heavy,
        ),
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
        impl.setup_dendropy_tree(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(3),
        hstrat.fixed_resolution_algo.Policy(5),
    ],
)
@pytest.mark.parametrize(
    "bias_adjustment",
    [
        None,
        "sample_ancestral_rollbacks",
        hstrat.UniformPrior(),
        hstrat.ArbitraryPrior(),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_determinism(orig_tree, retention_policy, bias_adjustment, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )

    first_reconst = hstrat.build_tree_trie(
        extant_population, bias_adjustment=bias_adjustment
    )
    for _rep in range(10):
        _ = _rep
        second_reconst = hstrat.build_tree_trie(
            [wrap(col) for col in extant_population],
            bias_adjustment=bias_adjustment,
        )
        assert first_reconst.equals(second_reconst)


@pytest.mark.parametrize(
    "orig_tree",
    [
        impl.setup_dendropy_tree(f"{assets_path}/nk_lexicaseselection.csv"),
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.recency_proportional_resolution_algo.Policy(4),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_reconstructed_taxon_labels(orig_tree, retention_policy, wrap):
    num_depositions = 10

    extant_population = hstrat.descend_template_phylogeny_dendropy(
        orig_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        ).CloneNthDescendant(num_depositions),
    )
    taxon_labels = [str(id(x)) for x in extant_population]

    reconst_df = hstrat.build_tree_trie(
        [*map(wrap, extant_population)],
        taxon_labels=taxon_labels,
    )
    assert "taxon_label" in reconst_df
    assert set(taxon_labels) < set(reconst_df["taxon_label"])

    reconst_df = hstrat.build_tree_trie(
        [*map(wrap, extant_population)],
    )
    assert "taxon_label" in reconst_df
    assert len(reconst_df["taxon_label"].unique()) == len(
        reconst_df["taxon_label"]
    )
    assert set(map(str, range(len(extant_population)))) < set(
        reconst_df["taxon_label"]
    )



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_does_definitively_have_no_common_ancestor_specimen(
    retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.does_definitively_have_no_common_ancestor(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column2)
    ) == hstrat.does_definitively_have_no_common_ancestor(column, column2)

    assert hstrat.does_definitively_have_no_common_ancestor(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column)
    ) == hstrat.does_definitively_have_no_common_ancestor(column, column)

    assert hstrat.does_definitively_have_no_common_ancestor(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(child1)
    ) == hstrat.does_definitively_have_no_common_ancestor(column, child1)

    assert hstrat.does_definitively_have_no_common_ancestor(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.does_definitively_have_no_common_ancestor(child1, child2)

    column2.DepositStrata(100)
    assert hstrat.does_definitively_have_no_common_ancestor(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(column2)
    ) == hstrat.does_definitively_have_no_common_ancestor(child1, column2)


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
def test_DefinitivelySharesNoCommonAncestorWith1(retention_policy):
    while True:
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        if hstrat.does_definitively_have_no_common_ancestor(c1, c2):
            assert hstrat.does_definitively_have_no_common_ancestor(c2, c1)
            for __ in range(100):
                c1.DepositStratum()
                assert hstrat.does_definitively_have_no_common_ancestor(c1, c2)
                assert hstrat.does_definitively_have_no_common_ancestor(c2, c1)
            for __ in range(100):
                c2.DepositStratum()
                assert hstrat.does_definitively_have_no_common_ancestor(c1, c2)
                assert hstrat.does_definitively_have_no_common_ancestor(c2, c1)
            break

    while True:
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        if not hstrat.does_definitively_have_no_common_ancestor(c1, c2):
            assert not hstrat.does_definitively_have_no_common_ancestor(c2, c1)
            for __ in range(100):
                c1.DepositStratum()
                assert not hstrat.does_definitively_have_no_common_ancestor(
                    c1, c2
                )
                assert not hstrat.does_definitively_have_no_common_ancestor(
                    c2, c1
                )
            for __ in range(100):
                c2.DepositStratum()
                assert not hstrat.does_definitively_have_no_common_ancestor(
                    c1, c2
                )
                assert not hstrat.does_definitively_have_no_common_ancestor(
                    c2, c1
                )
            break

    for rep in range(100):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        assert hstrat.does_definitively_have_no_common_ancestor(c1, c2)
        assert hstrat.does_definitively_have_no_common_ancestor(c2, c1)


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
def test_DefinitivelySharesNoCommonAncestorWith2(retention_policy):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=retention_policy,
    )
    c2 = c1.Clone()
    assert not hstrat.does_definitively_have_no_common_ancestor(c1, c2)
    assert not hstrat.does_definitively_have_no_common_ancestor(c2, c1)
    for __ in range(100):
        c1.DepositStratum()
        assert not hstrat.does_definitively_have_no_common_ancestor(c1, c2)
        assert not hstrat.does_definitively_have_no_common_ancestor(c2, c1)
    for __ in range(100):
        c2.DepositStratum()
        assert not hstrat.does_definitively_have_no_common_ancestor(c1, c2)
        assert not hstrat.does_definitively_have_no_common_ancestor(c2, c1)

    for rep in range(100):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        c2 = c1.Clone()
        assert not hstrat.does_definitively_have_no_common_ancestor(c1, c2)
        assert not hstrat.does_definitively_have_no_common_ancestor(c2, c1)


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.does_definitively_have_no_common_ancestor(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.does_definitively_have_no_common_ancestor(a, b)



import typing

import opytional as opyt

from ...._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    argsort,
    give_len,
)
from ._TrieInnerNode import TrieInnerNode


def build_trie_from_artifacts(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    taxon_labels: typing.Optional[typing.Iterable],
    progress_wrap: typing.Callable,
) -> TrieInnerNode:
    """Implementation detail for build_tree_trie_ensemble.

    See `build_tree_trie` for parameter descriptions.
    """
    taxon_labels = list(
        opyt.or_value(
            taxon_labels,
            map(str, range(len(population))),
        )
    )

    root = TrieInnerNode(rank=None, differentia=None)

    is_perfectly_synchronous = all(
        artifact.GetNumStrataDeposited()
        == population[0].GetNumStrataDeposited()
        for artifact in population
    )

    sort_order = argsort([x.GetNumStrataDeposited() for x in population])
    sorted_labels = [taxon_labels[i] for i in sort_order]
    sorted_population = [population[i] for i in sort_order]
    for label, artifact in progress_wrap(
        give_len(zip(sorted_labels, sorted_population), len(population))
    ):

        if is_perfectly_synchronous:
            root.InsertTaxon(label, artifact.IterRankDifferentiaZip())
        else:
            res = root.GetDeepestCongruousAlleleOrigination(
                artifact.IterRankDifferentiaZip(copyable=True)
            )
            node, subsequent_allele_genesis_iter = res
            node.InsertTaxon(label, subsequent_allele_genesis_iter)

    return root



from hstrat._auxiliary_lib import deep_listify, generate_omission_subsets


def test_generate_omission_subsets_with_list():
    assert deep_listify(generate_omission_subsets([1, 2, 3])) == [
        [2, 3],
        [1, 3],
        [1, 2],
    ]


def test_generate_omission_subsets_with_tuple():
    assert deep_listify(generate_omission_subsets((4, 5, 6))) == [
        [5, 6],
        [4, 6],
        [4, 5],
    ]


def test_generate_omission_subsets_with_string():
    assert deep_listify(generate_omission_subsets("ab")) == [["b"], ["a"]]


def test_generate_omission_subsets_with_empty_sequence():
    assert deep_listify(generate_omission_subsets([])) == []



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_DiffRetainedRanks(ordered_store, wrap):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
    )
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )

    assert hstrat.diff_retained_ranks(wrap(first), wrap(second)) == (
        set(),
        set(),
    )
    assert hstrat.diff_retained_ranks(wrap(second), wrap(first)) == (
        set(),
        set(),
    )

    first.DepositStratum()
    assert hstrat.diff_retained_ranks(wrap(first), wrap(second)) == (
        {1},
        set(),
    )
    assert hstrat.diff_retained_ranks(wrap(second), wrap(first)) == (
        set(),
        {1},
    )

    second.DepositStratum()
    assert hstrat.diff_retained_ranks(wrap(first), wrap(second)) == (
        set(),
        set(),
    )
    assert hstrat.diff_retained_ranks(wrap(second), wrap(first)) == (
        set(),
        set(),
    )

    first.DepositStratum()
    assert hstrat.diff_retained_ranks(wrap(first), wrap(second)) == ({2}, {1})
    assert hstrat.diff_retained_ranks(wrap(second), wrap(first)) == ({1}, {2})

    second.DepositStratum()
    assert hstrat.diff_retained_ranks(wrap(first), wrap(second)) == (
        set(),
        {1},
    )
    assert hstrat.diff_retained_ranks(wrap(second), wrap(first)) == (
        {1},
        set(),
    )


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.diff_retained_ranks(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.diff_retained_ranks(a, b)



import numpy as np

from ._jit import jit


@jit(nopython=True)
def count_unique(array: np.ndarray) -> int:
    """How many unique values are contained in `array`?"""
    unique_values = set(array)
    return len(unique_values)



import pytest

from hstrat._auxiliary_lib import flat_len


def test_flat_len_not_iterable():
    result = flat_len(11)
    assert result == 1

    result = flat_len(None)
    assert result == 1

    result = flat_len(False)
    assert result == 1


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_empty_list(wrap):
    result = flat_len(wrap([]))
    assert result == 0


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_empty_list_nested(wrap):
    result = flat_len(wrap([wrap([])]))
    assert result == 0


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_flat_list_singleton(wrap):
    result = flat_len(wrap([1]))
    assert result == 1


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_flat_list(wrap):
    result = flat_len(wrap([1, 2, 3]))
    assert result == 3


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_nested_list(wrap):
    result = flat_len(wrap([[1, 2], [3, 4]]))
    assert result == 4


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_nested_tuple(wrap):
    result = flat_len(wrap([(1, 2), (3, 4)]))
    assert result == 4


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_mixed_nested(wrap):
    result = flat_len(wrap([1, wrap([2, wrap((3, 4))]), 5]))
    assert result == 5


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_iterator(wrap):
    result = flat_len(wrap(iter([1, 2, 3])))
    assert result == 3


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_generator(wrap):
    result = flat_len(wrap((x for x in range(3))))
    assert result == 3


@pytest.mark.parametrize("wrap", [iter, lambda x: x])
def test_flat_len_singleton_list_nested(wrap):
    result = flat_len(wrap([wrap(["asdf"])]))
    assert result == 4



import dataclasses
import typing


@dataclasses.dataclass(slots=True)
class SearchtableRecord:
    """Implementation detail for Searchtable."""

    rank: int
    taxon_id: int  # corresponds to alife standard `id`
    taxon_label: str

    ancestor_id: int = 0  # represents parent in the build tree
    differentia: typing.Optional[int] = None
    search_first_child_id: int = 0  # helper for downward traversal
    search_next_sibling_id: int = 0  # helper for downward traversal
    search_ancestor_id: int = 0  # parent in the search tree



import typing

import numpy as np

from ..._auxiliary_lib import pairwise
from ..priors._detail import PriorBase


# tried a numpy-based implementation but it was slower
# https://gist.github.com/mmore500/fc4caba18a98b20fbda3fb223cf49552
def estimate_rank_of_mrca_unbiased(
    coincident_ranks_from_first_disparity_through_first_commonality: typing.Iterator[
        int
    ],
    p_differentia_collision: float,
    prior: PriorBase,
) -> typing.Optional[float]:
    """Compute an estimate for the rank of the most recent common ancestor (MRCA) that, on average, avoids systematic over- or under-estimation.

    Parameters
    ----------
    coincident_ranks_from_first_disparity_through_first_commonality : typing.Iterator[int]
        Iterator of integer values indicating the coincident ranks between two taxa.

        Inclusive to first disparity and first commonality.
    p_differentia_collision : float
        The multiplicative inverse of the number of possible differentia.
    prior : PriorBase
        Prior expectation for the distribution of MRCA generation
        between hereditary stratigraphic columns.

    Returns
    -------
    typing.Optional[float]
        Estimated rank of the MRCA, or None if the two hereditary stratigraphic
        artifacts definitively share no common ancestor.
    """

    base = p_differentia_collision

    weights, expected_ranks, sum_weight = [], [], 0.0
    for (
        num_spurious_collisions,
        (
            end_exclusive,
            begin_inclusive,
        ),
    ) in enumerate(
        pairwise(
            coincident_ranks_from_first_disparity_through_first_commonality
        )
    ):
        expected_ranks.append(
            prior.CalcIntervalConditionedMean(begin_inclusive, end_exclusive)
        )
        weights.append(
            base**num_spurious_collisions
            * prior.CalcIntervalProbabilityProxy(
                begin_inclusive, end_exclusive
            )
        )
        sum_weight += weights[-1]

        # if whole rest of the remaining record cannot meaningfully impact
        # average, exit early
        if (
            begin_inclusive
            and base ** (num_spurious_collisions + 1)
            * prior.CalcIntervalProbabilityProxy(
                0,
                begin_inclusive,
            )
            * begin_inclusive
            < sum_weight * 10e-6  # numpy uses reltol 10e-5
        ):
            break

    try:
        return np.average(
            np.array(expected_ranks),
            weights=np.array(weights),
        )
    except ZeroDivisionError:
        # handle empty array case
        return None



import typing

import opytional as opyt

from ..stratum_retention_algorithms._detail import PolicySpecBase
from ._PropertyAtLeastParameterizer import PropertyAtLeastParameterizer


class PropertyExactlyParameterizer:
    """Parameterizes so evaluated property exactly equals a target value."""

    _impl: PropertyAtLeastParameterizer

    def __init__(
        self: "PropertyExactlyParameterizer",
        *args,
        **kwargs,
    ) -> None:
        """Init functor with parameterization requirements.

        Parameters
        ----------
        target_value : float or int
            The threshold property value to parameterize the policy to meet.
        policy_evaluator : int
            Functor that evaluates a policy at a particular parameter value.
        param_lower_bound : int
            Lower value on the range of parameter values to search, inclusive.
        param_upper_bound : int, optional
            Upper bound on the range of parameter values to search, inclusive.
            If None, no upper bound is used.
        """
        self._impl = PropertyAtLeastParameterizer(*args, **kwargs)

    def __call__(
        self: "PropertyExactlyParameterizer",
        policy_t: typing.Type,
    ) -> typing.Optional[PolicySpecBase]:
        """Solve for policy spec satisfying parameterization requirements."""
        policy_factory = self._impl._policy_evaluator._policy_param_focalizer(
            policy_t,
        )
        try:
            res = self._impl._try_calc_parameter(policy_t)
            if res is not None:
                assert self._impl._param_lower_bound <= res
            if res is not None and self._impl._param_upper_bound is not None:
                assert res <= self._impl._param_upper_bound
        except (MemoryError, OverflowError, RecursionError):
            return None

        if (
            res is not None
            and self._impl._target_value
            == self._impl._policy_evaluator(
                policy_t,
                res,
            )
        ):
            return policy_factory(res).GetSpec()
        else:
            return None

    def __repr__(self: "PropertyExactlyParameterizer") -> str:
        return f"""{
            PropertyExactlyParameterizer.__qualname__
        }(target_value={
            self._impl._target_value
        !r}, policy_evaluator={
            self._impl._policy_evaluator
        !r}, param_lower_bound={
            self._impl._param_lower_bound
        !r}, param_upper_bound={
            self._impl._param_upper_bound
        !r})"""

    def __str__(self: "PropertyExactlyParameterizer") -> str:
        title = "Exactly Parameterizer"
        return f"""{
            title
        } (target value: {
            self._impl._target_value
        }, evaluator: {
            self._impl._policy_evaluator
        }, param lower bound: {
            self._impl._param_lower_bound
        }, param upper bound: {
            opyt.or_value(self._impl._param_upper_bound, 'inf')
        })"""



import itertools as it
import random

from iterify import cyclify, iterify
import opytional as opyt
import pytest
from scipy import stats

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95, 0.88],
)
def test_calc_ranks_since_mrca_bounds_with_specimen(
    retention_policy, differentia_width, confidence_level
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_ranks_since_mrca_bounds_with(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_ranks_since_mrca_bounds_with(
        column, column2, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_ranks_since_mrca_bounds_with(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_ranks_since_mrca_bounds_with(
        column, column, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_ranks_since_mrca_bounds_with(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_ranks_since_mrca_bounds_with(
        column, child1, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_ranks_since_mrca_bounds_with(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_ranks_since_mrca_bounds_with(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )

    child1.DepositStrata(10)
    assert hstrat.calc_ranks_since_mrca_bounds_with(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_ranks_since_mrca_bounds_with(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_CalcRanksSinceMrcaBoundsWith(
    retention_policy,
    ordered_store,
):
    def make_bundle():
        return hstrat.HereditaryStratigraphicColumnBundle(
            {
                "test": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_ordered_store=ordered_store,
                    stratum_retention_policy=retention_policy,
                ),
                "control": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_ordered_store=ordered_store,
                    stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                ),
            }
        )

    column = make_bundle()
    frozen_copy = column.Clone()
    frozen_unrelated = make_bundle()
    population = [column.Clone() for __ in range(10)]
    forked_isolated = column.Clone()
    unrelated_isolated = make_bundle()

    for _generation in range(100):
        _ = _generation
        for f, s in it.chain(
            it.combinations(population, 2),
            zip(population, cyclify(forked_isolated)),
            zip(population, cyclify(frozen_copy)),
            zip(cyclify(forked_isolated), population),
            zip(cyclify(frozen_copy), population),
        ):
            lb, ub = hstrat.calc_ranks_since_mrca_bounds_with(
                f["test"],
                s["test"],
                prior="arbitrary",
            )
            actual_rank_of_mrca = hstrat.get_last_common_stratum_between(
                f["control"],
                s["control"],
            ).GetAnnotation()
            actual_ranks_since_mrca = (
                f.GetNumStrataDeposited() - actual_rank_of_mrca - 1
            )
            assert lb <= actual_ranks_since_mrca < ub

        for f, s in it.chain(
            zip(population, cyclify(frozen_unrelated)),
            zip(population, cyclify(unrelated_isolated)),
            zip(cyclify(frozen_unrelated), population),
            zip(cyclify(unrelated_isolated), population),
        ):
            assert (
                hstrat.calc_ranks_since_mrca_bounds_with(
                    f["test"],
                    s["test"],
                    prior="arbitrary",
                )
                is None
            )

        # advance generation
        random.shuffle(population)
        for target in range(3):
            population[target] = population[-1].Clone()
            population[target].DepositStratum(
                annotation=population[target].GetNumStrataDeposited(),
            )
        for individual in it.chain(
            iter(population),
            iterify(forked_isolated),
            iterify(unrelated_isolated),
        ):
            if random.choice([True, False]):
                individual.DepositStratum(
                    annotation=individual.GetNumStrataDeposited(),
                )


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            ) == hstrat.calc_ranks_since_mrca_bounds_with(
                second, first, prior="arbitrary"
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_validity(retention_policy, ordered_store):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for generation in range(100):
        for first, second in it.combinations(population, 2):
            rslcw = hstrat.calc_ranks_since_last_retained_commonality_with(
                first, second
            )
            if rslcw is not None:
                assert 0 <= rslcw <= generation

            rsfdw = hstrat.calc_ranks_since_first_retained_disparity_with(
                first, second
            )
            if rsfdw is not None:
                assert -1 <= rsfdw <= generation

            assert hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            ) is None or hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            ) == (
                opyt.or_value(rsfdw, -1) + 1,
                rslcw + 1,
            )
            if rslcw is not None and rsfdw is not None:
                assert rsfdw < rslcw

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95],
)
def test_CalcRanksSinceMrcaBoundsWith_narrow_shallow(
    retention_policy,
    differentia_width,
    confidence_level,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    steps = list(
        range(
            columns[
                0
            ].CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=1 - confidence_level,
            )
            - columns[0].GetNumStrataDeposited()
        )
    )

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        for c1, c2 in zip(column1, column2):
            assert (
                hstrat.calc_ranks_since_mrca_bounds_with(
                    c1,
                    c2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_ranks_since_mrca_bounds_with(
                    c2,
                    c1,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [100],
)
def test_CalcRanksSinceMrcaBoundsWith_narrow_with_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
):

    columns = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )
        for __ in range(20)
    ]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [col.Clone() for col in columns]
        column2 = [col.Clone() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        num_inside_bounds = 0
        num_outside_bounds = 0
        for c1, c2 in zip(column1, column2):
            res = hstrat.calc_ranks_since_mrca_bounds_with(
                c1,
                c2,
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is None:
                num_outside_bounds += 1
                continue

            lb, ub = res
            assert lb < ub
            assert lb >= 0
            assert ub >= 0

            num_inside_bounds += (
                lb <= c1.GetNumStrataDeposited() - 1 - mrca_rank < ub
            )
            num_outside_bounds += not (
                lb <= c1.GetNumStrataDeposited() - 1 - mrca_rank < ub
            )

            assert c1.GetNumStrataDeposited() - 1 - mrca_rank >= lb

        num_trials = num_inside_bounds + num_outside_bounds
        assert (
            0.001
            < stats.binom.cdf(
                n=num_trials,
                p=1 - confidence_level,
                k=num_outside_bounds,
            )
            < 0.999
        )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95],
)
@pytest.mark.parametrize(
    "mrca_rank",
    [100],
)
def test_CalcRanksSinceMrcaBoundsWith_narrow_no_mrca(
    retention_policy,
    differentia_width,
    confidence_level,
    mrca_rank,
):
    def make_column():
        return hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
            stratum_ordered_store=hstrat.HereditaryStratumOrderedStoreDict,
        )

    columns = [make_column() for __ in range(20)]

    for column in columns:
        column.DepositStrata(mrca_rank)

    steps = (0, 16, 51)

    for step1, step2 in it.product(steps, steps):
        column1 = [make_column() for col in columns]
        column2 = [make_column() for col in columns]
        for __ in range(step1):
            for col in column1:
                col.DepositStratum()
        for i in range(step2):
            for col in column2:
                col.DepositStratum()

        num_inside_bounds = 0
        num_outside_bounds = 0
        for c1, c2 in zip(column1, column2):
            res = hstrat.calc_ranks_since_mrca_bounds_with(
                c1,
                c2,
                prior="arbitrary",
                confidence_level=confidence_level,
            )

            if res is None:
                num_inside_bounds += 1
                continue

            lb, ub = res
            assert lb < ub
            assert lb >= 0
            assert ub >= 0

            num_outside_bounds += 1

        num_trials = num_inside_bounds + num_outside_bounds
        assert (
            0.001
            < stats.binom.cdf(
                n=num_trials,
                p=1 - confidence_level,
                k=num_outside_bounds,
            )
            < 0.999
        )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary"],
)
def test_artifact_types_equiv(differentia_width, policy, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_ranks_since_mrca_bounds_with(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            prior,
        ) == hstrat.calc_ranks_since_mrca_bounds_with(a, b, prior)



import networkx as nx

from hstrat._auxiliary_lib import random_tree


def test_random_tree():
    """Tests that a random tree is in fact a tree."""
    T = random_tree(10, seed=1234)
    assert nx.is_tree(T)


def test_random_directed_tree():
    """Generates a directed tree."""
    T = random_tree(10, seed=1234, create_using=nx.DiGraph())
    assert T.is_directed()


def test_random_tree_using_generator():
    """Tests that creating a random tree with a generator works"""
    G = nx.Graph()
    T = random_tree(10, seed=1234, create_using=G)
    assert nx.is_tree(T)


# NetworkX is distributed with the 3-clause BSD license.

# ::

#    Copyright (C) 2004-2024, NetworkX Developers
#    Aric Hagberg <hagberg@lanl.gov>
#    Dan Schult <dschult@colgate.edu>
#    Pieter Swart <swart@lanl.gov>
#    All rights reserved.

#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:

#      * Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above
#        copyright notice, this list of conditions and the following
#        disclaimer in the documentation and/or other materials provided
#        with the distribution.

#      * Neither the name of the NetworkX Developers nor the names of its
#        contributors may be used to endorse or promote products derived
#        from this software without specific prior written permission.

#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



import polars as pl
from tqdm import tqdm

from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)
from hstrat.phylogenetic_inference.tree._impl._build_tree_searchtable_cpp_impl_stub import (
    build_tree_searchtable_cpp_from_exploded,
)


def test_regression_original():
    exploded = pl.DataFrame(
        {
            "dstream_data_id": pl.Series(
                "dstream_data_id",
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
                dtype=pl.UInt64,
            ),
            "dstream_T": pl.Series(
                "dstream_T",
                [8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11],
                dtype=pl.UInt64,
            ),
            "dstream_value_bitwidth": pl.Series(
                "dstream_value_bitwidth",
                [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
                dtype=pl.UInt32,
            ),
            "dstream_value": pl.Series(
                "dstream_value",
                [0, 1, 3, 7, 2, 5, 4, 6, 0, 1, 3, 7, 2, 5, 9, 6],
                dtype=pl.UInt64,
            ),
            "dstream_Tbar": pl.Series(
                "dstream_Tbar",
                [0, 1, 3, 7, 2, 5, 4, 6, 0, 1, 3, 7, 2, 5, 9, 6],
                dtype=pl.UInt64,
            ),
        },
    ).select(
        pl.col(
            "dstream_data_id",
            "dstream_T",
            "dstream_Tbar",
            "dstream_value",
        )
        .sort_by("dstream_Tbar")
        .over(partition_by="dstream_data_id"),
    )
    res = build_tree_searchtable_cpp_from_exploded(
        exploded["dstream_data_id"].to_numpy(),
        exploded["dstream_T"].to_numpy(),
        exploded["dstream_Tbar"].to_numpy(),
        exploded["dstream_value"].to_numpy(),
        tqdm,
    )
    res = pl.DataFrame(res)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert alifestd_is_chronologically_ordered(
        alifestd_try_add_ancestor_list_col(
            res.with_columns(origin_time=pl.col("rank")).to_pandas(),
        ),
    )


def test_regression_distilled():
    exploded = pl.DataFrame(
        {
            "dstream_data_id": pl.Series(
                "dstream_data_id",
                [0, 0, 1],
                dtype=pl.UInt64,
            ),
            "dstream_T": pl.Series(
                "dstream_T",
                [8, 8, 11],
                dtype=pl.UInt64,
            ),
            "dstream_value_bitwidth": pl.Series(
                "dstream_value_bitwidth",
                [
                    8,
                    8,
                    8,
                ],
                dtype=pl.UInt32,
            ),
            "dstream_value": pl.Series(
                "dstream_value",
                [7, 6, 6],
                dtype=pl.UInt64,
            ),
            "dstream_Tbar": pl.Series(
                "dstream_Tbar",
                [7, 6, 6],
                dtype=pl.UInt64,
            ),
        },
    ).select(
        pl.col(
            "dstream_data_id",
            "dstream_T",
            "dstream_Tbar",
            "dstream_value",
        )
        .sort_by("dstream_Tbar")
        .over(partition_by="dstream_data_id"),
    )
    res = build_tree_searchtable_cpp_from_exploded(
        exploded["dstream_data_id"].to_numpy(),
        exploded["dstream_T"].to_numpy(),
        exploded["dstream_Tbar"].to_numpy(),
        exploded["dstream_value"].to_numpy(),
        tqdm,
    )
    res = pl.DataFrame(res)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert alifestd_is_chronologically_ordered(
        alifestd_try_add_ancestor_list_col(
            res.with_columns(origin_time=pl.col("rank")).to_pandas(),
        ),
    )



import itertools as it
import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_earliest_detectable_mrca_among import (
    calc_rank_of_earliest_detectable_mrca_among,
)
from ._calc_rank_of_mrca_bounds_among import calc_rank_of_mrca_bounds_among


def does_share_any_common_ancestor(
    population: typing.Iterable[HereditaryStratigraphicArtifact],
    confidence_level: float = 0.95,
) -> typing.Optional[bool]:
    """Determine if common ancestry is evidenced within the population.

    If insufficient common ranks between strata are available to
    resolve any common ancestor, returns None.

    Note that stratum rention policies are strictly required to permanently
    retain the most ancient stratum.

    Parameters
    ----------
    confidence_level : float, optional
        The probability that we will correctly conclude no common ancestor
        is shared if, indeed, no common ancestor is actually shared. Default
        0.95.

    See Also
    --------
    does_definitively_share_no_common_ancestor :
        Can we definitively conclude that first and second share no common
        ancestor?
    """
    pop_tee1, pop_tee2 = it.tee(population)

    if (
        calc_rank_of_earliest_detectable_mrca_among(
            pop_tee1,
            confidence_level=confidence_level,
        )
        is None
    ):
        return None

    mrca_bounds = calc_rank_of_mrca_bounds_among(
        pop_tee2, "arbitrary", confidence_level
    )
    return mrca_bounds is not None



import typing

import pandas as pd
import polars as pl
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import coerce_to_pandas
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.NopTriePostprocessor,
        detail.TriePostprocessorBase,
    )


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_nop_mutate(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 0, 0],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
        },
    )
    df_ = hstrat.NopTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=False
    )
    assert df is not df_
    pd.testing.assert_frame_equal(coerce_to_pandas(df), coerce_to_pandas(df_))

    df = hstrat.NopTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=True
    )
    pd.testing.assert_frame_equal(coerce_to_pandas(df), coerce_to_pandas(df_))



from ._HereditaryStratumOrderedStoreDict import (
    HereditaryStratumOrderedStoreDict,
)
from ._HereditaryStratumOrderedStoreList import (
    HereditaryStratumOrderedStoreList,
)
from ._HereditaryStratumOrderedStoreTree import (
    HereditaryStratumOrderedStoreTree,
)

provided_stratum_ordered_stores = [
    HereditaryStratumOrderedStoreDict,
    HereditaryStratumOrderedStoreList,
    HereditaryStratumOrderedStoreTree,
]



import typing

from .._auxiliary_lib import bit_drop_msb
from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_from_packet import col_from_packet
from ._impl import DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH


def col_from_int(
    value: int,
    differentia_bit_width: int,
    stratum_retention_policy: typing.Callable,
    differentiae_byte_bit_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_width: int = (
        DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
    ),
    value_byte_width: typing.Optional[int] = None,
) -> HereditaryStratigraphicColumn:
    """Deserialize a `HereditaryStratigraphicColumn` from an integer
    representation.

    Integer representation is packet binary representation plus a sentry bit
    at the most significant bit position. Sentry bit prevents loss of leading
    zero bits. If `value_byte_width` is not `None`, an appropriate sentry bit
    is added to the value if it is not already present.

    Assumes big endian byte order.
    """

    if value_byte_width is not None:
        assert value.bit_length() <= value_byte_width * 8
        value |= 1 << (value_byte_width * 8)  # sentry bit

    if value.bit_length() % 8 != 1:  # sentry bit
        raise ValueError(
            f"Invalid integer representation {value} {bin(value)}. "
            "Must be an even byte multiple plus sentry bit. "
            "Probably missing sentry bit.",
        )
    semantic_value = bit_drop_msb(value)  # drop sentry bit
    buffer = semantic_value.to_bytes(
        (value.bit_length() - 1) // 8, byteorder="big", signed=False
    )
    return col_from_packet(
        packet=buffer,
        differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=stratum_retention_policy,
        differentiae_byte_bit_order=differentiae_byte_bit_order,
        num_strata_deposited_byte_order=num_strata_deposited_byte_order,
        num_strata_deposited_byte_width=num_strata_deposited_byte_width,
    )



import typing

from ..._detail import PolicyCouplerBase
from ..._impl import CalcMrcaUncertaintyRelUpperBoundWorstCase
from .._PolicySpec import PolicySpec


class CalcMrcaUncertaintyRelUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
        actual_rank_of_mrca: int,
    ) -> float:
        """At most, how much relative uncertainty to estimate rank of MRCA? Inclusive."""
        # rectify negative-indexed actual_rank_of_mrca
        if actual_rank_of_mrca is not None and actual_rank_of_mrca < 0:
            least_last_rank = min(
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
            actual_rank_of_mrca += least_last_rank
            assert actual_rank_of_mrca >= 0

        if (
            first_num_strata_deposited <= 2
            or second_num_strata_deposited <= 2
            or actual_rank_of_mrca
            in (
                first_num_strata_deposited - 1,
                second_num_strata_deposited - 1,
            )
        ):
            return 0.0

        abs_upper_bound = policy.CalcMrcaUncertaintyAbsUpperBound(
            first_num_strata_deposited,
            second_num_strata_deposited,
            actual_rank_of_mrca,
        )

        least_last_rank = min(
            first_num_strata_deposited - 1,
            second_num_strata_deposited - 1,
        )
        least_recency = (
            least_last_rank - actual_rank_of_mrca
            if actual_rank_of_mrca is not None
            else 1
        )

        res = abs_upper_bound / least_recency

        # tighten to worst-possible case given number of strata deposited
        return min(
            res,
            CalcMrcaUncertaintyRelUpperBoundWorstCase()(
                policy,
                first_num_strata_deposited,
                second_num_strata_deposited,
                actual_rank_of_mrca,
            ),
        )



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    hash_salt: int

    def __init__(self: "PolicySpec", hash_salt: int) -> None:
        """Construct the policy spec.

        Parameters
        ----------
        hash_salt : int
            Salt value fed into hash used to deterministically choose whether
            to keep or drop ranks.
        """
        self._hash_salt = hash_salt

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (self._hash_salt,) == (
            other._hash_salt,
        )

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(hash_salt={
            self._hash_salt
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (hash salt: {
            self._hash_salt
        })"""

    def GetHashSalt(self: "PolicySpec") -> int:
        return self._hash_salt

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return "Pseudostochastic Stratum Retention Algorithm"



from collections import defaultdict
from functools import lru_cache
import typing

import more_itertools as mit

from ..._auxiliary_lib import demark
from ...genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratum,
    HereditaryStratumOrderedStoreList,
)


def _calc_node_depths(
    descending_tree_iterable: typing.Iterable,
    get_parent: typing.Callable[[typing.Any], typing.Any],
    get_stem_length: typing.Callable[[typing.Any], int],
    demark: typing.Callable[[typing.Any], typing.Hashable] = demark,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.Dict[int, int]:
    """Descend tree to prepare lookup table of `demark(node)` -> node depth."""

    node_depth_lookup = dict()

    descending_tree_iterator = iter(progress_wrap(descending_tree_iterable))

    for root_node in descending_tree_iterator:
        node_depth_lookup[demark(root_node)] = 0
        break

    for node in descending_tree_iterator:
        stem_length = get_stem_length(node)
        parent_node = get_parent(node)
        node_depth_lookup[demark(node)] = (
            node_depth_lookup[demark(parent_node)] + stem_length
        )

    return node_depth_lookup


def _educe_stratum_ordered_store(
    ascending_lineage_iterable: typing.Iterable,
    deposition_count_lookup: typing.Dict[int, int],
    stem_strata_lookup: typing.Dict[
        int, typing.Callable[[int], HereditaryStratum]
    ],
    stratum_retention_policy: typing.Any,
    demark: typing.Callable[[typing.Any], typing.Hashable] = demark,
) -> HereditaryStratumOrderedStoreList:
    """Prepare strata required by one extant lineage member, using cache lookup
    to ensure that identical strata are provided where common ancestry is
    shared with previously processed extant lineage members."""

    try:
        extant_node = ascending_lineage_iterable[0]
    except TypeError:
        # usage ensures ascending_lineage_iterable not empty
        (extant_node,), ascending_lineage_iterable = mit.spy(
            ascending_lineage_iterable
        )

    extant_deposition_count = deposition_count_lookup[demark(extant_node)]

    rising_required_ranks_iterator = (
        stratum_retention_policy.IterRetainedRanks(
            extant_deposition_count,
        )
    )

    # pairwise ensures we exclude root node
    stratum_ordered_store = HereditaryStratumOrderedStoreList()
    try:
        descending_lineage_iterator = iter(
            reversed(ascending_lineage_iterable)
        )
    except TypeError:
        descending_lineage_iterator = iter(
            reversed([*ascending_lineage_iterable])
        )

    cur_node = next(descending_lineage_iterator)

    for rank in rising_required_ranks_iterator:
        while rank >= deposition_count_lookup[demark(cur_node)]:
            cur_node = next(descending_lineage_iterator)

        rank_stratum_lookup = stem_strata_lookup[demark(cur_node)]
        stratum_ordered_store.DepositStratum(
            rank=rank,
            stratum=rank_stratum_lookup(rank),
        )

    return stratum_ordered_store, extant_deposition_count


def descend_template_phylogeny_posthoc(
    ascending_lineage_iterables: typing.Iterable[typing.Iterable],
    descending_tree_iterable: typing.Iterable,
    get_parent: typing.Callable[[typing.Any], typing.Any],
    get_stem_length: typing.Callable[[typing.Any], int],
    seed_column: HereditaryStratigraphicColumn,
    demark: typing.Callable[[typing.Any], typing.Hashable] = demark,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Generate a population of hereditary stratigraphic columns that could
    have resulted from the template phylogeny.

    Calculates required ranks according to stratum retention policy for each
    extant lineage member, creating returned columns with strata at those
    ranks. Uses cache lookup to ensure that identical strata are provided where
    extant lineage members share common ancestry.

    Requires `seed_column`'s stratum retention policy to provide
    `IterRetainedRanks()` method.

    Prefer to use `descend_template_phylogeny`, which will automatically
    delegate to posthoc implementation if stratum retention policy requirements
    are met. See `descend_template_phylogeny` for parameter and return value details.
    """

    stratum_retention_policy = seed_column._stratum_retention_policy
    assert stratum_retention_policy.IterRetainedRanks is not None

    tree_depth_lookup = _calc_node_depths(
        descending_tree_iterable,
        get_parent,
        get_stem_length,
        progress_wrap=progress_wrap,
        demark=demark,
    )
    deposition_count_lookup = {
        k: v + seed_column.GetNumStrataDeposited()
        for k, v in tree_depth_lookup.items()
    }

    stem_strata_lookup = defaultdict(
        # use lru_cache as defaultdict with default factory conditioned on key
        lambda: lru_cache(maxsize=None)(
            lambda rank: (
                # if applicable, use stratum from seed column
                # otherwise, create new stratum
                seed_column.GetStratumAtRank(rank)
                if rank < seed_column.GetNumStrataDeposited()
                else seed_column._CreateStratum(rank)
            )
        )
    )

    extant_population = [
        HereditaryStratigraphicColumn(
            always_store_rank_in_stratum=seed_column._always_store_rank_in_stratum,
            stratum_retention_policy=stratum_retention_policy,
            stratum_differentia_bit_width=seed_column.GetStratumDifferentiaBitWidth(),
            stratum_ordered_store=_educe_stratum_ordered_store(
                iter_,
                deposition_count_lookup,
                stem_strata_lookup,
                stratum_retention_policy,
                demark=demark,
            ),
        )
        for iter_ in progress_wrap(ascending_lineage_iterables)
    ]
    return extant_population



import pandas as pd

from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_id_col import alifestd_make_ancestor_id_col


def alifestd_try_add_ancestor_id_col(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add an ancestor_id column to the input DataFrame if the phylogeny is
    asexual and the column does not already exist.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if alifestd_is_asexual(phylogeny_df) and "ancestor_id" not in phylogeny_df:
        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        phylogeny_df["ancestor_id"] = alifestd_make_ancestor_id_col(
            phylogeny_df["id"], phylogeny_df["ancestor_list"]
        )

    return phylogeny_df



import typing

from .._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)


def calc_ranks_since_first_retained_disparity_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """Determine generations since divergence with particular confidence.

    How many depositions have elapsed along the focal column's line of
    descent since the first mismatching strata at the same rank between
    focal and other?

    Returns
    -------
    int, optional
        The number of depositions elapsed or None if no disparity (i.e.,
        both columns have same number of strata deposited and the most
        recent stratum is common between focal and other).

    Parameters
    ----------
    confidence_level : float, optional
        With what probability should the true number of ranks since the
        first disparity be less than or equal to the returned estimate?
        Default 0.95.

    Notes
    -----
    Returns -1 if focal and other share no mismatching strata at common
    ranks but other has more strata deposited than focal.

    The true number of ranks since the first disparity with other is
    guaranteed strictly less than or equal to the returned estimate when
    confidence_level < 0.5.
    """
    first_disparate_rank = calc_rank_of_first_retained_disparity_between(
        focal,
        other,
        confidence_level=confidence_level,
    )
    if first_disparate_rank is None:
        return None
    else:
        assert focal.GetNumStrataDeposited()
        res = focal.GetNumStrataDeposited() - 1 - first_disparate_rank
        assert -1 <= res < focal.GetNumStrataDeposited()
        return res



from hstrat import hstrat
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.CompoundTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_empty():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor([])(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert not hasattr(leaf, "origin_time")
    assert not hasattr(inner, "origin_time")
    assert not hasattr(root, "origin_time")
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")


def test_singleton():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(),
        ]
    )(root, p_differentia_collision=0.5, mutate=True)
    assert getattr(leaf, "origin_time") == 0
    assert getattr(inner, "origin_time") == 0
    assert getattr(root, "origin_time") == 0
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")


def test_pair():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(),
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(
                assigned_property="blueberry"
            ),
        ]
    )(root, p_differentia_collision=0.5, mutate=True)
    assert getattr(leaf, "origin_time") == 0
    assert getattr(inner, "origin_time") == 0
    assert getattr(root, "origin_time") == 0
    assert getattr(leaf, "blueberry") == 0
    assert getattr(inner, "blueberry") == 0
    assert getattr(root, "blueberry") == 0


def test_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")

    processed_root = hstrat.CompoundTriePostprocessor(
        [
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(),
            hstrat.AssignOriginTimeNodeRankTriePostprocessor(
                assigned_property="blueberry"
            ),
        ]
    )(root, p_differentia_collision=0.5, mutate=False)
    assert not hasattr(leaf, "origin_time")
    assert not hasattr(inner, "origin_time")
    assert not hasattr(root, "origin_time")
    assert not hasattr(leaf, "blueberry")
    assert not hasattr(inner, "blueberry")
    assert not hasattr(root, "blueberry")

    assert hasattr(processed_root, "origin_time")
    assert hasattr(processed_root, "blueberry")



import copy
import typing


def col_records_from_pop_records(
    pop_records: typing.Dict,
    mutate: bool = False,
) -> typing.Iterator[typing.Dict]:
    """Generate column records from a dictionary of builtin types representing
    a serialized population of `HereditaryStratigraphicColumn`s.

    Parameters
    ----------
    pop_records : dict
        A dictionary containing the serialized population data. It must have a
        'columns' field, which is a list of dictionaries representing the
        columns.

        Other common fields will be propagated into individual column entries.
    mutate : bool, default False
        Are side effects on the input argument `pop_records` allowed?

    Yields
    ------
    dict
        Dictionariew of builtin types representing individual column records.
    """
    col_records = pop_records["columns"]

    if not mutate:
        col_records = copy.deepcopy(col_records)

    for col_record in col_records:
        for common_field in (
            "policy",
            "policy_algo",
            "policy_spec",
            "differentia_bit_width",
            "hstrat_version",
        ):
            col_record[common_field] = pop_records[common_field]
        yield col_record



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_make_empty,
    alifestd_splay_polytomies,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)


def test_alifestd_splay_polytomies_empty():
    pd.testing.assert_frame_equal(
        alifestd_splay_polytomies(alifestd_make_empty()).reset_index(
            drop=True
        ),
        alifestd_try_add_ancestor_id_col(alifestd_make_empty()).reset_index(
            drop=True
        ),
    )


def test_alifestd_splay_polytomies_singleton():
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "ancestor_id": [0],
            "trait": [1.2],
        }
    )
    expected_df = df.copy()
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize("shuffle", [True, False])
def test_alifestd_splay_polytomies_no_polytomies(shuffle):
    # Test case for a phylogeny dataframe with only bifurcations
    #    99
    #  /   \
    # 2     3
    #     /   \
    #    4     5
    df = pd.DataFrame(
        {
            "id": [99, 2, 3, 4, 5],
            "ancestor_list": ["[none]", "[99]", "[99]", "[3]", "[3]"],
            "ancestor_id": [99, 99, 99, 3, 3],
            "trait": [1.2, 3.1, 2.5, 1.7, 0.9],
        }
    )
    if shuffle:
        df = df.sample(frac=1).reset_index(drop=True)
    expected_df = df.copy()
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize("shuffle", [True, False])
def test_alifestd_splay_polytomies_no_polytomies_unifurcation(shuffle):
    # Test case for a phylogeny dataframe with no polytomies,
    # but with unifurcations
    #    0
    #    |
    #    1
    #  /   \
    # 2     4
    # |     |
    # 5     7
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4, 5, 7],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[1]",
                "[2]",
                "[4]",
            ],
            "ancestor_id": [0, 0, 1, 1, 2, 4],
            "trait": [1.2, 3.1, 2.5, 0.9, 12.1, 9.9],
        }
    )
    if shuffle:
        df = df.sample(frac=1).reset_index(drop=True)
    expected_df = df.copy()
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize("shuffle", [True, False])
def test_alifestd_splay_polytomies_no_polytomies_only_unifurcation(shuffle):
    # Test case for a phylogeny dataframe with only unifurcations
    #    0
    #    |
    #    1
    #    |
    #    2
    #    |
    #    3
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[2]",
            ],
            "ancestor_id": [0, 0, 1, 2],
            "trait": [1.2, 3.1, 2.5, 0.9],
        }
    )
    if shuffle:
        df = df.sample(frac=1).reset_index(drop=True)
    expected_df = df.copy()
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize("rotate", [True, False])
def test_alifestd_splay_polytomies_one_polytomy_of_size_three(rotate):
    # Test case for a phylogeny dataframe with one polytomy of size three
    #   1
    #  /|\
    # 2 3 4
    df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4],
            "ancestor_list": ["[none]", "[1]", "[1]", "[1]"],
            "ancestor_id": [1, 1, 1, 1],
            "trait": [1.2, 3.1, 2.5, 1.7],
            "origin_time": [0, 1, 1, 1],
        }
    )
    if rotate:
        df = pd.concat([df.iloc[1:], df.iloc[:1]]).reset_index(drop=True)
        assert len(df) == 4
    #   1
    #  / \
    # 2   5
    #    /  \
    #   3    4
    expected_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5],
            "ancestor_list": ["[none]", "[1]", "[5]", "[5]", "[1]"],
            "ancestor_id": [1, 1, 5, 5, 1],
            "trait": [1.2, 3.1, 2.5, 1.7, 1.2],
            "origin_time": [0, 1, 1, 1, 0],
        }
    )
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(
        result_df.sort_values("id").reset_index(drop=True),
        expected_df.sort_values("id").reset_index(drop=True),
    )


def test_alifestd_splay_polytomies_one_polytomy_of_size_three_with_unifurcation():
    # Test case for a phylogeny dataframe with one polytomy of size three
    # and unifurcations
    #    1
    #  / | \
    # 2  0  3
    # |     |
    # 4     5
    #       |
    #       6
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [
                "[1]",
                "[none]",
                "[1]",
                "[1]",
                "[2]",
                "[3]",
                "[5]",
            ],
            "ancestor_id": [1, 1, 1, 1, 2, 3, 5],
            "trait": [1.2, 3.1, 2.5, 1.7, 0.9, 101.9, 22.0],
        }
    )
    #    1
    #  /   \
    # 0     7
    #     /   \
    #    2     3
    #    |     |
    #    4     5
    #          |
    #          6
    expected_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7],
            "ancestor_list": [
                "[1]",
                "[none]",
                "[7]",
                "[7]",
                "[2]",
                "[3]",
                "[5]",
                "[1]",
            ],
            "ancestor_id": [1, 1, 7, 7, 2, 3, 5, 1],
            "trait": [1.2, 3.1, 2.5, 1.7, 0.9, 101.9, 22.0, 3.1],
        }
    )
    result_df = alifestd_splay_polytomies(df, mutate=False)
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)

    result_df = alifestd_splay_polytomies(
        df.reindex([0, 1, 2, 4, 3, 5, 6]), mutate=False
    )
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(
        result_df,
        expected_df.reindex([0, 1, 2, 4, 3, 5, 6, 7]).reset_index(drop=True),
    )


def test_alifestd_splay_polytomies_two_polytomies_with_unifurcation():
    # Test case for a phylogeny dataframe with two polytomies, one of size
    # three and one of size four, and a unifurcation
    #    0
    #    | \
    #    1  10
    #  / | \
    # 2  3  4
    # |    /||\
    # 5   / || \
    #    / / | |
    #   6 7  8 9
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[1]",
                "[1]",
                "[2]",
                "[4]",
                "[4]",
                "[4]",
                "[4]",
                "[0]",
            ],
            "ancestor_id": [0, 0, 1, 1, 1, 2, 4, 4, 4, 4, 0],
        }
    )
    #    0
    #    | \
    #    1  10
    #  /   \
    # 2     11
    # |    /  \
    # 5   3    4
    #         / \
    #        6   12
    #           /  \
    #          7    13
    #              /  \
    #             8    9
    expected_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
            "ancestor_list": [
                "[none]",
                "[0]",
                "[1]",
                "[11]",
                "[11]",
                "[2]",
                "[4]",
                "[12]",
                "[13]",
                "[13]",
                "[0]",
                "[1]",
                "[4]",
                "[12]",
            ],
            "ancestor_id": [0, 0, 1, 11, 11, 2, 4, 12, 13, 13, 0, 1, 4, 12],
        }
    )
    result_df = alifestd_assign_contiguous_ids(
        alifestd_splay_polytomies(df, mutate=False)
    )
    alifestd_validate(result_df)
    pd.testing.assert_frame_equal(result_df, expected_df)


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            {
                "id": [99, 2, 3, 4, 5],
                "ancestor_list": ["[none]", "[99]", "[99]", "[3]", "[3]"],
                "ancestor_id": [99, 99, 99, 3, 3],
                "trait": [1.2, 3.1, 2.5, 1.7, 0.9],
            }
        ),  #  <- no polytomies
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5, 6],
                "ancestor_list": [
                    "[1]",
                    "[none]",
                    "[1]",
                    "[1]",
                    "[2]",
                    "[3]",
                    "[5]",
                ],
                "ancestor_id": [1, 1, 1, 1, 2, 3, 5],
                "trait": [1.2, 3.1, 2.5, 1.7, 0.9, 101.9, 22.0],
            }
        ),  #  <- has polytomy
    ],
)
def test_alifestd_splay_polytomies_mutate(df):
    df_ = df.copy()
    alifestd_splay_polytomies(df, mutate=False)
    pd.testing.assert_frame_equal(df_, df)



import typing

from ._with_omission import with_omission


def generate_omission_subsets(
    sequence: typing.Sequence,
) -> typing.Iterator[typing.Iterator]:
    """Generate all possible single-omission subsets of an input sequence, with
    the kth element omitted from the kth subset.

    Parameters
    ----------
    sequence : sequence
        The input sequence to generate subsets from.

    Yields
    ------
    iterable
        An iterable of the input sequence with one element omitted.

    Examples
    --------
    >>> list(generate_omission_subsets([1, 2, 3]))
    [[2, 3], [1, 3], [1, 2]]

    >>> list(generate_omission_subsets('abc'))
    [['b', 'c'], ['a', 'c'], ['a', 'b']]
    """
    for index, __ in enumerate(sequence):
        yield with_omission(sequence, index)



import textwrap

from ._collapse_nonleading_whitespace import collapse_nonleading_whitespace
from ._join_paragraphs_from_one_sentence_per_line import (
    join_paragraphs_from_one_sentence_per_line,
)
from ._textwrap_respect_indents import textwrap_respect_indents


def format_cli_description(raw_message: str) -> str:
    """Fix whitespace to pretty-print description message on CLI, for use
    with argparse's `description` argument.

    See `hstrat.dataframe.surface_unpack_reconstruct for usage example.
    """
    message = raw_message
    message = join_paragraphs_from_one_sentence_per_line(message)
    message = collapse_nonleading_whitespace(message)
    message = textwrap_respect_indents(message)
    message = textwrap.indent(message, prefix="  ")
    return message



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_try_add_ancestor_id_col,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
                )[0:1],
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
                )[0:1],
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        )[-1:0],
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        )[0:1],
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny-empty-list-notation.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny-noncompact1.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny-noncompact2.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny-uniq.csv"
        ),
    ],
)
@pytest.mark.parametrize(
    "apply_combine",
    [
        lambda x: alifestd_aggregate_phylogenies(
            [
                x,
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
                ),
            ]
        ),
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1),
        lambda x: alifestd_try_add_ancestor_id_col(x.sample(frac=1)),
        lambda x: x,
    ],
)
def test_valid_asexual(phylogeny_df, apply_combine, apply):
    phylogeny_df = apply(apply_combine(phylogeny_df.copy()))

    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_validate(phylogeny_df)
    assert phylogeny_df_.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_path",
    [
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-missing-id-column.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-ancestor_list-syntax1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-ancestor_list-syntax2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-ancestor_list-syntax3.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-ancestor_list-syntax4.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-duplicate-id1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-duplicate-id2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-duplicate-id3.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-mismatched-ancestor_id-ancestor_list-columns1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-mismatched-ancestor_id-ancestor_list-columns2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-mismatched-ancestor_id-ancestor_list-columns3.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-missing-ancestor_list-column1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-missing-ancestor_list-column2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-negative-id1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-negative-id2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-negative-id3.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-nonexistant-ancestor_id-id1.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-nonexistant-ancestor_id-id2.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-nonexistant-ancestor_id-id3.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-nonexistant-ancestor_id-id4.csv",
        f"{assets_path}/example-standard-toy-asexual-phylogeny--invalid-nonexistant-ancestor_list-id.csv",
    ],
)
@pytest.mark.parametrize(
    "apply_combine",
    [
        lambda x: pd.concat(
            [
                x,
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny-uniq.csv"
                ),
            ]
        ),
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "apply",
    [
        alifestd_try_add_ancestor_id_col,
        lambda x: x.sample(frac=1),
        lambda x: alifestd_try_add_ancestor_id_col(x.sample(frac=1)),
        lambda x: x,
    ],
)
def test_invalid_asexual(phylogeny_path, apply_combine, apply):
    phylogeny_df = apply_combine(pd.read_csv(phylogeny_path))
    try:
        phylogeny_df = apply(phylogeny_df)
    except Exception:
        pass

    phylogeny_df_ = phylogeny_df.copy()
    assert not alifestd_validate(phylogeny_df)
    assert phylogeny_df_.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        )[0:2],
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny-empty-list-notation.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny-noncompact1.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny-noncompact2.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny-uniq.csv"
        ),
    ],
)
@pytest.mark.parametrize(
    "apply_combine",
    [
        lambda x: alifestd_aggregate_phylogenies(
            [
                x,
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
                ),
            ]
        ),
        lambda x: alifestd_aggregate_phylogenies(
            [
                x,
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
            ]
        ),
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "apply_shuffle",
    [
        lambda x: x.sample(frac=1),
        lambda x: x,
    ],
)
def test_valid_sexual(phylogeny_df, apply_combine, apply_shuffle):
    phylogeny_df = apply_shuffle(apply_combine(phylogeny_df.copy()))

    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_validate(phylogeny_df)
    assert phylogeny_df_.equals(phylogeny_df)


@pytest.mark.parametrize(
    "phylogeny_path",
    [
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-missing-id-column.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-ancestor_list-syntax1.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-ancestor_list-syntax2.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-ancestor_list-syntax3.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-ancestor_list-syntax4.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-duplicate-id1.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-duplicate-id2.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-duplicate-id3.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-mismatched-ancestor_id-ancestor_list-columns1.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-mismatched-ancestor_id-ancestor_list-columns2.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-negative-id1.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-negative-id2.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-negative-id3.csv",
        f"{assets_path}/example-standard-toy-sexual-phylogeny--invalid-nonexistant-ancestor_list-id.csv",
    ],
)
@pytest.mark.parametrize(
    "apply_combine",
    [
        lambda x: pd.concat(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-asexual-phylogeny-uniq.csv"
                ),
                x,
            ]
        ),
        lambda x: pd.concat(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny-uniq.csv"
                ),
                x,
            ]
        ),
        lambda x: x,
    ],
)
@pytest.mark.parametrize(
    "apply_shuffle",
    [
        lambda x: x.sample(frac=1),
        lambda x: x,
    ],
)
def test_invalid_sexual(phylogeny_path, apply_combine, apply_shuffle):
    phylogeny_df = apply_shuffle(apply_combine(pd.read_csv(phylogeny_path)))

    phylogeny_df_ = phylogeny_df.copy()
    assert not alifestd_validate(phylogeny_df)
    assert phylogeny_df_.equals(phylogeny_df)


def test_empty():
    empty1 = pd.DataFrame(
        {
            "ancestor_list": [],
            "id": [],
        }
    )
    assert alifestd_validate(empty1)

    empty2 = pd.DataFrame(
        {
            "ancestor_list": [],
            "id": [],
            "origin_time": [],
        }
    )
    assert alifestd_validate(empty2)



import typing


# adapted from http://stackoverflow.com/questions/3071415/
def argsort(seq: typing.Sequence, reverse: bool = False) -> typing.List[int]:
    """Return the indices that would sort a sequence in ascending order.

    Parameters
    ----------
    seq : sequence
        Input sequence to be sorted.
    reverse : bool, optional
        If True, the indices that would sort the sequence in descending order
        are returned instead.
        Default is False.

    Returns
    -------
    indices : list of int
        List of indices that would sort the sequence.

    Notes
    -----
    This function is analogous to numpy.argsort.

    Examples
    --------
    >>> seq = [4.0, 2.0, 1.0, 3.0]
    >>> argsort(seq)
    [2, 1, 3, 0]
    >>> argsort(seq, reverse=True)
    [0, 3, 1, 2]
    """
    return sorted(range(len(seq)), key=seq.__getitem__, reverse=reverse)



import typing

import typing_extensions

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_from_packet import col_from_packet
from ._impl import DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH


def col_from_packet_buffer(
    packet_buffer: typing_extensions.Buffer,
    differentia_bit_width: int,
    stratum_retention_policy: typing.Callable,
    differentiae_byte_bit_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_width: int = (
        DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
    ),
) -> HereditaryStratigraphicColumn:
    """Deserialize a `HereditaryStratigraphicColumn` from a buffer containing
    the differentia packet at the front, then stored differentia values.

    Use when buffer size exceeds packet size.

    See Also
    --------
    col_from_packet: use when buffer size equals packet size.
    """

    num_strata_deposited = int.from_bytes(
        packet_buffer[:num_strata_deposited_byte_width],
        byteorder=num_strata_deposited_byte_order,
        signed=False,
    )
    num_strata_retained = stratum_retention_policy.CalcNumStrataRetainedExact(
        num_strata_deposited,
    )
    packet_num_bits = (
        num_strata_deposited_byte_width * 8
        + num_strata_retained * differentia_bit_width
    )
    packet_num_bytes = (packet_num_bits + 7) // 8  # +7 rounds up
    assert packet_num_bytes <= len(packet_buffer)
    return col_from_packet(
        packet=packet_buffer[:packet_num_bytes],
        differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=stratum_retention_policy,
        differentiae_byte_bit_order=differentiae_byte_bit_order,
        num_strata_deposited_byte_order=num_strata_deposited_byte_order,
        num_strata_deposited_byte_width=num_strata_deposited_byte_width,
    )



from ._render_to_numeral_system import render_to_numeral_system


def render_to_base64url(n: int) -> str:
    """Convert an integer to its digit representation in the base64url encoding.

    Parameters
    ----------
    n : int
        The non-negative integer to be converted to a base64url representation.

    Returns
    -------
    str
        A string of characters representing the base64url encoding of the given
        integer.

    Raises
    ------
    AssertionError
        If the integer is negative or if the base64url alphabet contains
        repeated characters.

    See Also
    --------
    render_to_numeral_system : A general function for converting integers to
    custom base representations.

    Notes
    -----
    The base64url encoding is specified in RFC 4648, and is similar to the
    standard base64 encoding, but replaces '+' and '/' with '-' and '_'. This
    is often used in web applications as a URL-safe encoding for binary data.

    Examples
    --------
    >>> render_to_base64url(163)
    'Cj'
    >>> render_to_base64url(42)
    'q'
    >>> render_to_base64url(7)
    'H'
    """
    base64url_alphabet = (
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_"
    )
    return render_to_numeral_system(n, base64url_alphabet)



import typing

import numpy as np
import pandas as pd
import polars as pl

from ...._auxiliary_lib import (
    AnyTreeFastPreOrderIter,
    anytree_iterative_deepcopy,
)
from ...priors._detail import PriorBase
from .._impl import TrieInnerNode, TrieLeafNode
from ._AssignOriginTimeNaiveTriePostprocessor import (
    AssignOriginTimeNaiveTriePostprocessor,
)
from ._detail import TriePostprocessorBase


def _call_anytree(
    ftor: "AssignOriginTimeExpectedValueTriePostprocessor",
    trie: TrieInnerNode,
    p_differentia_collision: float,
    progress_wrap: typing.Callable,
) -> TrieInnerNode:
    """Implementation detail."""
    trie = AssignOriginTimeNaiveTriePostprocessor(
        prior=ftor._prior, assigned_property="_naive_origin_time"
    )(
        trie,
        p_differentia_collision=p_differentia_collision,
        mutate=True,
        progress_wrap=progress_wrap,
    )

    for node in progress_wrap(AnyTreeFastPreOrderIter(trie)):
        if node.is_leaf:
            setattr(node, ftor._assigned_property, node.rank)
            assert isinstance(node, TrieLeafNode)
        elif node.parent is None:
            setattr(node, ftor._assigned_property, 0)
        else:
            assert node.children
            weights = (
                p_differentia_collision,
                1
                * ftor._prior.CalcIntervalProbabilityProxy(
                    node.rank,
                    min(
                        (
                            child.rank
                            for child in node.children
                            if not child.is_leaf
                        ),
                        default=node.rank + 1,
                    ),
                ),
            )
            values = (
                getattr(node.parent, ftor._assigned_property),
                node._naive_origin_time,
            )
            collision_corrected_origin_time = np.average(
                values,
                weights=weights,
            )
            setattr(
                node,
                ftor._assigned_property,
                collision_corrected_origin_time,
            )

    return trie


class AssignOriginTimeExpectedValueTriePostprocessor(TriePostprocessorBase):
    """Functor to assign origin time property to trie nodes using expected
    values over the distribution of possible differentia collisions.

    Computes the origin time of trie nodes based on the expected value, taking
    into account the probability of differentia collision and the prior
    distribution expected for ancestor origin times.
    """

    _assigned_property: str  # property name for assigned origin time
    _prior: PriorBase  # prior expectation for ancestor origin times

    def __init__(
        self: "AssignOriginTimeExpectedValueTriePostprocessor",
        prior: PriorBase,
        assigned_property: str = "origin_time",
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        prior : PriorBase
            The prior distribution used to calculate expected values.
        assigned_property : str, default "origin_time"
            The property name for the assigned origin time.
        """
        self._assigned_property = assigned_property
        self._prior = prior

    def __call__(
        self: "AssignOriginTimeExpectedValueTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Assign origin times to trie nodes.

        Parameters
        ----------
        trie : TrieInnerNode
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie with assigned origin times.
        """
        if isinstance(trie, TrieInnerNode):
            if not mutate:
                trie = anytree_iterative_deepcopy(
                    trie, progress_wrap=progress_wrap
                )
            return _call_anytree(
                self,
                trie,
                p_differentia_collision,
                progress_wrap=progress_wrap,
            )
        elif isinstance(trie, (pl.DataFrame, pd.DataFrame)):
            raise NotImplementedError  # pragma: no cover
        else:
            raise TypeError  # pragma: no cover



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcMrcaUncertaintyRelUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelUpperBound",
        policy: typing.Optional[PolicyCouplerBase],
        first_num_strata_deposited: typing.Optional[int],
        second_num_strata_deposited: typing.Optional[int],
        actual_rank_of_mrca: typing.Optional[int],
    ) -> float:
        """At most, how much relative uncertainty to estimate rank of MRCA?
        Inclusive."""
        return 0.0



import typing

from ..._auxiliary_lib import demark
from ...genome_instrumentation import HereditaryStratigraphicColumn


def descend_template_phylogeny_naive(
    ascending_lineage_iterables: typing.Iterable[typing.Iterable],
    descending_tree_iterable: typing.Iterable,
    get_parent: typing.Callable[[typing.Any], typing.Any],
    get_stem_length: typing.Callable[[typing.Any], int],
    seed_column: HereditaryStratigraphicColumn,
    demark: typing.Callable[[typing.Any], typing.Hashable] = demark,
    progress_wrap: typing.Callable = lambda x: x,
) -> typing.List[HereditaryStratigraphicColumn]:
    """Generate a population of hereditary stratigraphic columns that could
    have resulted from the template phylogeny.

    Traverses phylogenetic tree in topological order, generating a clone column
    with `get_stem_length(node)` additional stratum deposits for each node
    (including internal nodes). Uses `CloneNthDescendant()` instead of `n`
    calls to `CloneDescendant()` to improve efficiency where
    `get_stem_length(node)` > 1.

    Prefer to use `descend_template_phylogeny`, which will automatically
    delegate between naive and posthoc implementation, unless performance
    considerations merit manual override.
    """

    hstrat_column_lookup = dict()  # node id -> column

    descending_tree_iterator = iter(descending_tree_iterable)

    for root_node in descending_tree_iterator:
        hstrat_column_lookup[demark(root_node)] = seed_column
        break

    for node in progress_wrap(descending_tree_iterator):
        stem_length = get_stem_length(node)
        parent_node = get_parent(node)
        parent_hstrat_column = hstrat_column_lookup[demark(parent_node)]

        node_hstrat_column = parent_hstrat_column.CloneNthDescendant(
            stem_length
        )
        hstrat_column_lookup[demark(node)] = node_hstrat_column

    extant_population = [
        # extant node
        hstrat_column_lookup[demark(next(iter(ascending_lineage_iterable)))]
        for ascending_lineage_iterable in ascending_lineage_iterables
    ]
    return extant_population
